/************************************************************************/
/* Plugin to deal with customised ASM files from SAM8 disassemblies     */
/************************************************************************/ 

#include <ida.hpp>
#include <idp.hpp>
#include <loader.hpp>
#include <fpro.h>
#include <name.hpp>
#include <bytes.hpp>
#include <ctype.h>


// processor code for SAM8
#define PLFM_SAM8 0x8020

// offset to code segment
#define SAM8_CODESEG_START 0
#define SAM8_CODESEG_SIZE 0x10000

// offset to external data segment
#define SAM8_EDATASEG_START 0x800000
#define SAM8_EDATASEG_SIZE 0x10000


/*
 * Save a line to the file
 *
 * @param fp File concerned
 * @param line Unprocessed line to save
 */
void save_line(FILE* fp, char* line) {
  char buf[256];

  // clear stuff from the string
  tag_remove(line, buf, 256);

  // output it
  qfprintf(fp, "%s\n", buf);
}


/*
 * Output EQU directives
 *
 * @param fp File concerned
 * @return Number of lines output
 */
int output_equs(FILE* fp) {
  char buf[256];
  int numLines = 0;
  
  // Now, output the EQUs
  ea_t curAddr = SAM8_EDATASEG_START;
  while((curAddr < (SAM8_EDATASEG_START + SAM8_EDATASEG_SIZE)) && 
        (!wasBreak())) {
    // try and get name at this address + output it
    char* curName = get_name(BADADDR, curAddr, buf, 256);
    if (curName != NULL) {
      qfprintf(fp, "%s EQU 0%xh\n", curName, 
               curAddr - SAM8_EDATASEG_START); numLines++;
      numLines++;
    }
    
    // show address to user
    if (!(curAddr & 0xff)) showAddr(curAddr);
    
    // move to next address
    curAddr++;
  }
  
  // return number of lines output
  return numLines;
}


/*
 * Generate the text of a disassembly
 * @param fp File concerned
 * @param EA1 Starting ea
 * @param EA2 Ending ea +1
 * @return Number of lines output
 */
int gen_text(FILE* fp, ea_t EA1, ea_t EA2) {
  int numLines = 0;

  // setup the linearray
  int dumping = 0;
  linearray_t lnar(&dumping);
  idaplace_t idaplace(EA1, 0);
  lnar.set_place(&idaplace);

  // loop until end of line array, or user cancelled
  while((!lnar.ending()) && (!wasBreak())) {
    // move to next item
    char* p = lnar.down();

    // skip the first text chunk on the line (the address)
    while(*p) {
      // found end of first line
      if (isspace(*p)) break;

      // next character
      p++;
    }
    if (isspace(*p)) p++; // skip one final space character if there is one

    // get the current place & display status to user
    idaplace_t* place = (idaplace_t*) lnar.get_place();
    ulong ea = place->ea;
    if ((ea & 0xFF) == 0) showAddr(ea);

    // if we're at ea 0, and the line contains the comment that we should
    // replace with the EQU directives...
    if ((ea == 0) && 
        (strstr(p, "; These will appear here when output using the samaout plugin"))) {
      // output the EQU definitions
      numLines += output_equs(fp);
    } else {
      // save current line
      save_line(fp, p);
      numLines++;
    }

    // hit the end yet?
    if (ea >= EA2) break;
  }

  // return number of generated lines
  return numLines;
}


/**
 * Entry point for plugin
 */
void idaapi plugin_main(int) {
  // get filename from user
  char* filename = askfile_c(1, "*.ASM", "Enter SAMA ASM filename for %s", 
                             get_root_filename());
  if (filename == NULL) {
    return;
  }

  // if the file already exists, ask the user what do do
  int operation = 1;
  if (qfileexist(filename)) {
    operation = askbuttons_c("~A~ppend", "~O~verwrite", "~C~ancel",
                             -1, 
                             "Output file '%s' already exists.", filename);
  }

  // perform appropriate operation
  FILE* fp = NULL;
  switch(operation) {
  case 1: // append
    fp = qfopen(filename, "a");
    break;
  
  case 0: // overwrite
    fp = qfopen(filename, "w");
    break;
  
  case -1: // cancel
    return;
  }

  // check if file was opened ok
  if (fp == NULL) {
    warning("Unable to open destination file '%s'.", filename);
    msg("Failed to output SAMA .ASM to '%s'.\n", filename);
    return;
  }

  // show the wait box
  show_wait_box("Creating the output file.");

  // generate the disassembly text
  int numLines = gen_text(fp, SAM8_CODESEG_START, 
                          SAM8_CODESEG_START + SAM8_CODESEG_SIZE);
  
  // close it again
  qfclose(fp);

  // Hide the wait box
  hide_wait_box();

  // Tell the user if it wasn't an abort
  if (!wasBreak) {
    msg("SAMA Assembler file created, total %i lines.\n", numLines);
  }
}


/**
 * Init the plugin (works out if plugin should be loaded)
 */
int idaapi init(void) {
  // refuse to load if we're not being used for a SAM8 processor
  if (ph.id != PLFM_SAM8) return PLUGIN_SKIP;

  // Let ourselves be loaded
  return PLUGIN_OK;
}

/**
 * Terminate the plugin
 */
void idaapi term(void) {
}



/************************************************************************/
/* Necessary strings for plugin                                         */
/************************************************************************/

/** Comment about the plugin */
char comment[] = "SAMA ASM generator";

/** Long comment about plugin */
char help[] =
"SAMA ASM generator\n"
"\n"
"This module allows .ASM files to be saved from SAM8 "
"disassemblies in SAMA format";

/** Preferred name of plugin; can be overridden in plugins.cfg */
char wanted_name[] = "Save SAMA .ASM file";

/** Preferred hotkey for plugin; can be overridden in plugins.cfg */
char wanted_hotkey[] = "Alt-Z";


/************************************************************************/
/* Plugin descriptor block                                              */
/************************************************************************/
extern "C" plugin_t PLUGIN =
{
  IDP_INTERFACE_VERSION,
  0,                    // plugin flags
  init,
  term,
  plugin_main,
  comment,
  help,
  wanted_name,
  wanted_hotkey
};
